package dev.fitko.fitconnect.client;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.model.SubmissionPayload;
import dev.fitko.fitconnect.client.steps.sender.*;
import dev.fitko.fitconnect.client.strategies.SendEncryptedSubmissionStrategy;
import dev.fitko.fitconnect.client.strategies.SendNewSubmissionStrategy;
import dev.fitko.fitconnect.client.strategies.SubmitStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import static dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType.APPLICATION_JSON;
import static dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType.APPLICATION_XML;

import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildDataPayload;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildEncryptedDataPayload;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildServiceType;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.toAttachmentPayloads;

/**
 * A fluent client for announcing and handing in a {@link SubmitSubmission}
 */
public final class SenderClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(SenderClient.class);

    private SenderClient() {
    }

    public static StartStep build(final Sender sender, final Environment environment) {
        return new ClientBuilder(sender, environment);
    }

    /**
     * ClientBuilder that implements all steps to guide through the API calls
     */
    public static class ClientBuilder implements
            StartStep,
            AttachmentsStep, EncryptedAttachmentsStep,
            DataStep, EncryptedDataStep,
            EncryptedMetadataStep,
            DestinationStep,
            ServiceTypeStep,
            SubmitStep {

        private final Sender sender;
        private final Environment environment;
        private SubmitStrategy submitStrategy;
        private SubmissionPayload submissionPayload;

        public ClientBuilder(final Sender sender, final Environment environment) {
            this.sender = sender;
            this.environment = environment;
        }

        @Override
        public AttachmentsStep newSubmission() {
            this.submitStrategy = new SendNewSubmissionStrategy(sender, environment);
            this.submissionPayload = new SubmissionPayload();
            return this;
        }

        @Override
        public EncryptedAttachmentsStep newSubmissionWithEncryptedData() {
            this.submitStrategy = new SendEncryptedSubmissionStrategy(sender);
            this.submissionPayload = new SubmissionPayload();
            return this;
        }

        @Override
        public Optional<String> getPublicKey(final UUID destinationId) {
            try {
                return retrievePublicKey(destinationId);
            } catch (final RestApiException | KeyNotRetrievedException e) {
                LOGGER.error("Public encryption key could not be retrieved", e);
            }
            return Optional.empty();
        }

        @Override
        public ServiceTypeStep withDestination(final UUID destinationId) {
            if (destinationId == null) {
                LOGGER.error("DestinationId was null");
                return this;
            }
            submissionPayload = submissionPayload.withDestinationId(destinationId);
            return this;
        }

        @Override
        public DataStep withAttachments(final List<File> attachments) {
            if (attachments == null) {
                LOGGER.error("Attachments were null");
                return this;
            }
            submissionPayload = submissionPayload.withAttachments(toAttachmentPayloads(attachments));
            return this;
        }

        @Override
        public DataStep withAttachment(final File attachment) {
            if (attachment == null) {
                LOGGER.error("Attachment was null");
                return this;
            }
            withAttachments(List.of(attachment));
            return this;
        }

        @Override
        public DestinationStep withJsonData(final String jsonData) {
            if (jsonData == null) {
                LOGGER.error("Json data was null");
                return this;
            }
            final ValidationResult validationResult = sender.validateJsonFormat(jsonData);
            if (validationResult.hasError()) {
                LOGGER.error("Data is not in expected json format, please provide valid json {}", validationResult.getError().getMessage());
                return this;
            }
            submissionPayload = submissionPayload.withData(buildDataPayload(jsonData, APPLICATION_JSON));
            return this;
        }


        @Override
        public DestinationStep withXmlData(final String xmlData) {
            if (xmlData == null) {
                LOGGER.error("Xml data was null");
                return this;
            }
            final ValidationResult validationResult = sender.validateXmlFormat(xmlData);
            if (validationResult.hasError()) {
                LOGGER.error("Data is not in expected xml format, please provide valid xml {}", validationResult.getError().getMessage());
                return this;
            }
            submissionPayload = submissionPayload.withData(buildDataPayload(xmlData, APPLICATION_XML));
            return this;
        }

        @Override
        public EncryptedDataStep withEncryptedAttachments(final Map<UUID, String> encryptedAttachments) {
            if (encryptedAttachments == null) {
                LOGGER.error("Encrypted attachments were null");
                return this;
            }
            submissionPayload = submissionPayload.withAttachments(toAttachmentPayloads(encryptedAttachments));
            return this;
        }

        @Override
        public EncryptedDataStep withEncryptedAttachment(final UUID attachmentId, final String encryptedAttachment) {
            if (attachmentId == null) {
                LOGGER.error("AttachmentId was null");
                return this;
            }
            if (encryptedAttachment == null) {
                LOGGER.error("EncryptedAttachment was null");
                return this;
            }
            return withEncryptedAttachments(Map.of(attachmentId, encryptedAttachment));
        }

        @Override
        public EncryptedMetadataStep withEncryptedJsonData(final String encryptedJsonData) {
            if (encryptedJsonData == null) {
                LOGGER.error("Encrypted json data was null");
                return this;
            }
            submissionPayload = submissionPayload.withData(buildEncryptedDataPayload(encryptedJsonData, APPLICATION_JSON));
            return this;
        }

        @Override
        public EncryptedMetadataStep withEncryptedXmlData(final String encryptedXmlData) {
            if (encryptedXmlData == null) {
                LOGGER.error("Encrypted xml data was null");
                return this;
            }
            submissionPayload = submissionPayload.withData(buildEncryptedDataPayload(encryptedXmlData, APPLICATION_XML));
            return this;
        }

        @Override
        public DestinationStep withEncryptedMetadata(final String encryptedMetadata) {
            if (encryptedMetadata == null) {
                LOGGER.error("Encrypted metadata was null");
                return this;
            }
            submissionPayload = submissionPayload.withEncryptedMetadata(encryptedMetadata);
            return this;
        }

        @Override
        public SubmitStep withServiceType(final String serviceTypeName, final String leikaKey) {
            if (serviceTypeName == null) {
                LOGGER.error("ServiceType name was null");
                return this;
            }
            if (leikaKey == null) {
                LOGGER.error("LeikaKey was null");
                return this;
            }
            submissionPayload = submissionPayload.withServiceType(buildServiceType(serviceTypeName, leikaKey));
            return this;
        }

        @Override
        public SubmitStep withServiceType(final String serviceTypeName, final String description, final String leikaKey) {
            if (serviceTypeName == null) {
                LOGGER.error("ServiceType name was null");
                return this;
            }
            if (description == null) {
                LOGGER.error("ServiceType description was null");
                return this;
            }
            if (leikaKey == null) {
                LOGGER.error("LeikaKey was null");
                return this;
            }
            submissionPayload = submissionPayload.withServiceType(buildServiceType(serviceTypeName, description, leikaKey));
            return this;
        }

        @Override
        public Optional<SubmitSubmission> submit() {
            return isValidSubmissionData() ? submitStrategy.send(submissionPayload) : Optional.empty();
        }

        private boolean isValidSubmissionData() {
            if (submissionPayload == null) {
                return false;
            } else if (submissionPayload.getDestinationId() == null) {
                LOGGER.error("DestinationId is mandatory, but was null.");
                return false;
            } else if (submissionPayload.getData() == null) {
                LOGGER.error("Data is mandatory, but was null.");
                return false;
            } else if (submissionPayload.getServiceType() == null) {
                LOGGER.error("ServiceType is mandatory, but was null.");
                return false;
            } else {
                return true;
            }
        }

        private Optional<String> retrievePublicKey(final UUID destinationId) {
            final Destination destination = sender.getDestination(destinationId);
            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId, destination.getEncryptionKid());
            final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
            if (validationResult.hasError() && !environment.isAllowInsecurePublicKey()) {
                LOGGER.error("Invalid public key", validationResult.getError());
                return Optional.empty();
            }
            return Optional.of(encryptionKey.toJSONString());
        }
    }
}
