package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.client.model.AttachmentPayload;
import dev.fitko.fitconnect.client.model.DataPayload;
import dev.fitko.fitconnect.client.model.ServiceTypePayload;
import dev.fitko.fitconnect.client.model.SubmissionPayload;
import dev.fitko.fitconnect.core.util.Strings;

import java.io.File;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;

public final class SubmissionUtil {

    private SubmissionUtil(){}

    public static CreateSubmission createSubmission(final UUID destinationId, final ServiceType serviceType, final List<UUID> attachmentIdsToAnnounce) {
        return CreateSubmission.builder()
                .destinationId(destinationId)
                .announcedAttachments(attachmentIdsToAnnounce)
                .serviceType(serviceType)
                .build();
    }

    public static ServiceType buildServiceType(final ServiceTypePayload serviceTypePayload) {
        final var serviceType = new ServiceType();
        serviceType.setName(serviceTypePayload.getName());
        serviceType.setIdentifier(serviceTypePayload.getLeikaKey());
        if (!Strings.isNullOrEmpty(serviceTypePayload.getDescription())) {
            serviceType.setDescription(serviceTypePayload.getDescription());
        }
        return serviceType;
    }

    public static ServiceType buildServiceType(final String serviceTypeName, final String leikaKey) {
        final var serviceType = new ServiceType();
        serviceType.setName(serviceTypeName);
        serviceType.setIdentifier(leikaKey);
        return serviceType;
    }

    public static ServiceType buildServiceType(final String serviceTypeName, final String description, final String leikaKey) {
        final var serviceType = new ServiceType();
        serviceType.setName(serviceTypeName);
        serviceType.setDescription(description);
        serviceType.setIdentifier(leikaKey);
        return serviceType;
    }

    public static PublicServiceType buildPublicServiceType(final ServiceType serviceType) {
        final var publicServiceType = new PublicServiceType();
        publicServiceType.setIdentifier(serviceType.getIdentifier());
        publicServiceType.setName(serviceType.getName());
        if (serviceType.getDescription() != null) {
            publicServiceType.setDescription(serviceType.getDescription());
        }
        return publicServiceType;
    }

    public static Data buildData(final DataPayload dataPayload) {
        final var hash = new Hash();
        hash.setContent(dataPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(dataPayload.getMimeType());
        submissionSchema.setSchemaUri(dataPayload.getSchemaUri());

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);
        return data;
    }

    public static Metadata buildMetadataToSend(final ServiceType serviceType, final List<AttachmentPayload> encryptedAttachments, final DataPayload encryptedDataPayload) {
        final Data dataToSend = buildData(encryptedDataPayload);
        final PublicServiceType publicServiceType = buildPublicServiceType(serviceType);
        final List<Attachment> attachmentMetadata = toAttachmentMetadata(encryptedAttachments);
        return buildMetadata(attachmentMetadata, dataToSend, publicServiceType);
    }

    public static Metadata buildMetadata(final List<Attachment> attachments, final Data data, final PublicServiceType publicServiceType) {
        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(attachments);
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);
        metadata.setPublicServiceType(publicServiceType);
        return metadata;
    }

    public static CreateSubmission buildSubmissionToAnnounce(final SubmissionPayload submissionPayload) {
        final ServiceType serviceType = submissionPayload.getServiceType();
        final List<UUID> attachmentIds = toAttachmentIds(submissionPayload.getAttachments());
        return createSubmission(submissionPayload.getDestinationId(), serviceType, attachmentIds);
    }

    public static SubmitSubmission buildSubmitSubmission(final UUID submissionId, final SubmissionPayload submissionPayload) {
        final SubmitSubmission submission = new SubmitSubmission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(submissionPayload.getData().getEncryptedData());
        submission.setEncryptedMetadata(submissionPayload.getEncryptedMetadata());
        return submission;
    }

    public static SubmitSubmission buildSubmitSubmission(final UUID submissionId, final String encryptedData, final String encryptedMetadata) {
        final SubmitSubmission submission = new SubmitSubmission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);
        return submission;
    }

    public static Attachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new Attachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFile().getName());
        attachment.setMimeType(attachmentPayload.getMimeType());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }

    public static List<UUID> toAttachmentIds(final List<AttachmentPayload> attachmentPayloads) {
        return attachmentPayloads.stream()
                .map(AttachmentPayload::getAttachmentId)
                .collect(Collectors.toList());
    }

    public static List<Attachment> toAttachmentMetadata(final List<AttachmentPayload> attachmentPayloads) {
        return attachmentPayloads.stream()
                .map(SubmissionUtil::toHashedAttachment)
                .collect(Collectors.toList());
    }

    public static Function<Map.Entry<UUID, String>, AttachmentPayload> getEncryptedAttachmentPayload() {
        return attachment -> AttachmentPayload.builder()
                .encryptedData(attachment.getValue())
                .attachmentId(attachment.getKey())
                .build();
    }

    public static Function<File, AttachmentPayload> getFileAttachmentPayload() {
        return file -> AttachmentPayload.builder()
                .file(file)
                .attachmentId(UUID.randomUUID())
                .build();
    }

    public static List<AttachmentPayload> toAttachmentPayloads(final Map<UUID, String> encryptedAttachments) {
        return encryptedAttachments.entrySet()
                .stream()
                .map(getEncryptedAttachmentPayload())
                .collect(Collectors.toList());
    }

    public static List<AttachmentPayload> toAttachmentPayloads(final List<File> attachments) {
        return attachments.stream()
                //.filter(file -> Files.notExists(Path.of(file.getPath())))
                .map(getFileAttachmentPayload())
                .collect(Collectors.toList());
    }

    public static DataPayload buildDataPayload(final String data, final MimeType mimeType) {
        return DataPayload.builder()
                .rawData(data.getBytes(StandardCharsets.UTF_8))
                .mimeType(mimeType)
                .build();
    }

    public static DataPayload buildEncryptedDataPayload(final String encryptedData, final MimeType mimeType) {
        return DataPayload.builder()
                .encryptedData(encryptedData)
                .mimeType(mimeType)
                .build();
    }

    public static CreateSubmission buildSubmissionToAnnounce(final UUID destinationId, final ServiceType serviceType, final List<AttachmentPayload> encryptedAttachments) {
        final List<UUID> attachmentIdsToAnnounce = toAttachmentIds(encryptedAttachments);
        return createSubmission(destinationId, serviceType, attachmentIdsToAnnounce);
    }

    public static List<SubmissionSchema> getSubmissionSchemasFromDestination(final Destination destination) {
        return destination.getServices().stream()
                .map(DestinationService::getSubmissionSchemas)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
    }
    public static Optional<URI> getSchemaUriForMimeType(final List<SubmissionSchema> submissionSchemas, final MimeType mimeType) {
        return submissionSchemas.stream()
                .filter(schema -> schema.getMimeType().equals(mimeType))
                .map(SubmissionSchema::getSchemaUri)
                .findFirst();
    }

    public static AttachmentWithData buildAttachmentWithData(Attachment attachment, String encryptedAttachment) {
        return AttachmentWithData.builder()
                .encryptedData(encryptedAttachment)
                .attachmentMetadata(attachment)
                .build();
    }

}
